//
//  HTMLStore.m
//  League Manager
//
//  Created by Michael Privat on 9/20/10.
//  Copyright (c) 2010 __MyCompanyName__. All rights reserved.
//

#import <CoreData/CoreData.h>

#import "HTMLStore.h"
#import "HTMLStoreParser.h"

@interface HTMLStore (private)
+ (void)writeMetadata:(NSDictionary*)metadata toURL:(NSURL*)url;
+ (NSString *)makeUUID;
@end

@interface HTMLStore (htmlStuff)
- (void)newTableForEntity:(NSEntityDescription*)entity inHtml:(NSMutableString*)html;
@end

@implementation HTMLStore
#pragma mark -
#pragma mark NSPersistentStore

- (NSString *)type {
  return [[self metadata] objectForKey:NSStoreTypeKey];
}

- (NSString *)identifier {
  return [[self metadata] objectForKey:NSStoreUUIDKey];
}

- (id)initWithPersistentStoreCoordinator:(NSPersistentStoreCoordinator *)coordinator configurationName:(NSString *)configurationName URL:(NSURL *)url options:(NSDictionary *)options {
  self = [super initWithPersistentStoreCoordinator:coordinator configurationName:configurationName URL:url options:options];
  
  NSDictionary *metadata = [HTMLStore metadataForPersistentStoreWithURL:[self URL] error:nil];
  [self setMetadata:metadata];
  
  nodeCacheRef = [[NSMutableDictionary dictionary] retain];
  
  return self;
}

+ (NSDictionary *)metadataForPersistentStoreWithURL:(NSURL *)url error:(NSError **)error {
  NSString *path = [[url relativePath] stringByAppendingString:@".plist"];
  
  if(! [[NSFileManager defaultManager] fileExistsAtPath:path]) {
    NSMutableDictionary *metadata = [NSMutableDictionary dictionary];
    [metadata setValue:@"HTMLStore" forKey:NSStoreTypeKey];
    [metadata setValue:[HTMLStore makeUUID] forKey:NSStoreUUIDKey];
    [HTMLStore writeMetadata:metadata toURL:url];
    [@"<html/>" writeToURL:url atomically:YES encoding:[NSString defaultCStringEncoding] error:nil];
    
    NSLog(@"Created new store at %@", path);    
  }
  
  return [NSDictionary dictionaryWithContentsOfFile:path];
}

#pragma mark -
#pragma mark HTMLStore

- (void)updateCacheNode:(NSAtomicStoreCacheNode *)node fromManagedObject:(NSManagedObject *)managedObject {
    NSEntityDescription *entity = managedObject.entity;
    
    NSDictionary *attributes = [entity attributesByName];
    for(NSString *name in [attributes allKeys]) {
        [node setValue:[managedObject valueForKey:name] forKey:name];
    }
    
    NSDictionary *relationships = [entity relationshipsByName];
    for(NSString *name in [relationships allKeys]) {
        id value = [managedObject valueForKey:name];
        if([[relationships objectForKey:name] isToMany]) {
            NSSet *set = (NSSet*)value;
            NSMutableSet *data = [NSMutableSet set];
            for(NSManagedObject *managedObject in set) {
                NSManagedObjectID *oid = [managedObject objectID];  
                id referenceID = [self referenceObjectForObjectID:oid];              
                NSAtomicStoreCacheNode* n = [self nodeForReferenceObject:referenceID andObjectID:oid];
                [data addObject:n];
            }
            [node setValue:data forKey:name];            
        }
        else {
            NSManagedObject *managedObject = (NSManagedObject*)value;
            NSManagedObjectID *oid = [managedObject objectID];  
            id referenceID = [self referenceObjectForObjectID:oid];              
            NSAtomicStoreCacheNode* n = [self nodeForReferenceObject:referenceID andObjectID:oid];            
            [node setValue:n forKey:name];
        }
    }
}

- (NSAtomicStoreCacheNode *)nodeForReferenceObject:(id)reference andObjectID:(NSManagedObjectID *)oid {
    NSAtomicStoreCacheNode *node = [nodeCacheRef objectForKey:reference];
    if(node == nil) {
        node = [[[NSAtomicStoreCacheNode alloc] initWithObjectID:oid] autorelease];
        [nodeCacheRef setObject:node forKey:reference];
    }
    
    return node;
}

#pragma mark -
#pragma mark NSAtomicStore

- (id)newReferenceObjectForManagedObject:(NSManagedObject *)managedObject {
    return [[[managedObject objectID] URIRepresentation] description];
    
//  NSString *uuid = [HTMLStore makeUUID];
//  [uuid retain];
//  return uuid;
}

- (NSAtomicStoreCacheNode *)newCacheNodeForManagedObject:(NSManagedObject *)managedObject {
  NSManagedObjectID *oid = [managedObject objectID];  
  id referenceID = [self referenceObjectForObjectID:oid];  
  
  NSAtomicStoreCacheNode* node = [self nodeForReferenceObject:referenceID andObjectID:oid];
  [self updateCacheNode:node fromManagedObject:managedObject];
  return node;
}

-(void)appendStyles:(NSMutableString*)html {
    [html appendString:@"<head>"];
	[html appendString:@"<style>"];
    [html appendString:@"body {background-color:#ddd}"];
    [html appendString:@"h1 { font-size:18; }"];
    [html appendString:@"table {border-collapse:collapse;}"];
    [html appendString:@"table, th, td { border: 1px solid black; padding:4px}"];
    [html appendString:@"th {text-align:left;background-color:#88f;color:#fff}"];
    [html appendString:@"td {background-color:#ddf}"];
	[html appendString:@"</style>"];
    [html appendString:@"</head>"];
}

- (BOOL)save:(NSError **)error {
  NSURL *url = [self URL];
  
  // First update the metadata
  [HTMLStore writeMetadata:[self metadata] toURL:url];

  NSSet *n = [self cacheNodes];
  NSMutableArray *nodesArray = [[NSMutableArray alloc] init];
  for(id item in n) {
    [nodesArray addObject:item];
  }
  
  // Group all the nodes by entity name
  NSArray *nodes = [nodesArray sortedArrayUsingComparator:(NSComparator)^(id obj1, id obj2){
    NSAtomicStoreCacheNode *n1 = (NSAtomicStoreCacheNode*)obj1;
    NSAtomicStoreCacheNode *n2 = (NSAtomicStoreCacheNode*)obj2;

    NSManagedObjectID *oid1 = [n1 objectID];
    NSEntityDescription *entity1 = [oid1 entity];

    NSManagedObjectID *oid2 = [n2 objectID];
    NSEntityDescription *entity2 = [oid2 entity];

    return [entity1.name caseInsensitiveCompare:entity2.name];
  }];
  
  // Produce the HTML
  
  NSMutableString *html = [[NSMutableString alloc] init];
  
  [html appendString:@"<html>"];
  [self appendStyles:html];
  [html appendString:@"<body>"];
  
  NSEntityDescription *currentTable = nil;  
  for(NSAtomicStoreCacheNode *node in nodes) {
    NSManagedObjectID *oid = [node objectID];
    NSEntityDescription *entity = [oid entity];

    if([entity.name compare:currentTable.name] != NSOrderedSame) {      // New table
      // If there was a table open, close it
      if(currentTable != nil) {
        [html appendString:@"</table>"];
      }

      [self newTableForEntity:entity inHtml:html];
      currentTable = entity;
    }
    
    // Write the data
    
    [html appendString:@"<tr>"];
    
    id referenceID = [self referenceObjectForObjectID:oid];
    [html appendFormat:@"<td><a name='%@'/>%@</td>", referenceID, referenceID];
    
    {   // Attributes
      NSDictionary *attributes = [entity attributesByName];
      NSAttributeDescription *key = nil;
      NSEnumerator *enumerator = [attributes objectEnumerator];
      while((key = [enumerator nextObject]) != nil) {
        NSString *value = [node valueForKey:key.name];
        if(value == nil) value = @"(null)";
        
        [html appendFormat:@"<td>%@</td>", value];
      }
    }

    {   // Relationships
      NSDictionary *relationships = [entity relationshipsByName];
      NSRelationshipDescription *key = nil;
      NSEnumerator *enumerator = [relationships objectEnumerator];
      while((key = [enumerator nextObject]) != nil) {
        id value = [node valueForKey:key.name];
        if(value == nil) {
          [html appendString:@"<td>(null)</td>"];
        }
        else if(![key isToMany]) {  // One-to-One
          NSManagedObjectID *oid = [(NSAtomicStoreCacheNode*)value objectID];
          id referenceID = [self referenceObjectForObjectID:oid];
          
          [html appendFormat:@"<td><a href='#%@'>%@</a></td>", referenceID, referenceID];
        }
        else {  // One-to-Many
          NSSet* set = (NSSet*)value;
          if([set count] == 0) {
            [html appendString:@"<td>(null)</td>"];
          }
          else {
            [html appendString:@"<td><ul>"];
            for(NSAtomicStoreCacheNode *item in set) {
              id referenceID = [self referenceObjectForObjectID:[item objectID]];
              [html appendFormat:@"<li><a href='#%@'>%@</a></li>", referenceID, referenceID];
            }
            [html appendString:@"</ul></td>"];
          }
        }
      }
    }
    
    [html appendString:@"</tr>"];
    
  }
  
  // Close the last table if needed
  if(currentTable != nil) {
    [html appendString:@"</table>"];
  }
  
  // Close up the HTML
  
  [html appendString:@"</body>"];
  [html appendString:@"</html>"];
  
  // Write the HTML to disk
  NSString *path = [url relativePath];
  [html writeToFile:path atomically:YES encoding:[NSString defaultCStringEncoding] error:error];

  [html release];
  [nodesArray release];
  
  return YES;
}

- (void)newTableForEntity:(NSEntityDescription*)entity inHtml:(NSMutableString*)html {
  [html appendFormat:@"<h1>%@</h1>", entity.name];
  [html appendString:@"<table>"];
  
  [html appendString:@"<tr>"];
  [html appendString:@"<th>id</th>"];
  
  // Write the table header
  {   // Attributes
    NSDictionary *attributes = [entity attributesByName];
    NSAttributeDescription *key = nil;
    NSEnumerator *enumerator = [attributes objectEnumerator];
    while((key = [enumerator nextObject]) != nil) {
      [html appendFormat:@"<th>%@</th>", key.name];
    }
  }
  {   // Relationships
    NSDictionary *relationships = [entity relationshipsByName];
    NSRelationshipDescription *key = nil;
    NSEnumerator *enumerator = [relationships objectEnumerator];
    while((key = [enumerator nextObject]) != nil) {
      [html appendFormat:@"<th>%@</th>", key.name];
    }
  }
  
  [html appendString:@"</tr>"];
}

- (BOOL)load:(NSError **)error {
    NSURL* url = [self URL];
    
    // See if this is a new store
    NSString *path = [url relativePath];
    if(! [[NSFileManager defaultManager] fileExistsAtPath:path]) {
        NSMutableSet *nodes = [NSMutableSet set];
        [self addCacheNodes:nodes];
        return YES;
    }
    
    NSXMLParser *xmlParser = [[NSXMLParser alloc] initWithContentsOfURL:url];
    
    HTMLStoreParser *parserDelegate = [[HTMLStoreParser alloc] initWithStore:self];  
    [xmlParser setDelegate:parserDelegate];  
    
    BOOL success = [xmlParser parse];
    
    // Register all the nodes
    [self addCacheNodes:parserDelegate.nodes];
    
    [parserDelegate release];
    
    if(!success) NSLog(@"Unable to parse document");
    
    return success;
}


#pragma mark -
#pragma mark HTMLStore

+ (void)writeMetadata:(NSDictionary *)metadata toURL:(NSURL *)url {
  NSString *path = [[url relativePath] stringByAppendingString:@".plist"];
  [metadata writeToFile:path atomically:YES];
}

+ (NSString *)makeUUID {    
  CFUUIDRef uuidRef = CFUUIDCreate(NULL);
  CFStringRef uuidStringRef = CFUUIDCreateString(NULL, uuidRef);
  CFRelease(uuidRef);
  NSString* uuid = [NSString stringWithString:(NSString *)uuidStringRef];
  CFRelease(uuidStringRef);
  return uuid;
}

- (void)dealloc {
  [nodeCacheRef release];
  
  [super dealloc];
}

@end
